@@ -9,7 +9,9 @@
 -- License: BSD-2-clause
 -- 
 -- Console regions are displayed near the bottom of the console, and can be
--- updated concurrently by threads. Any other output displayed using
+-- updated concurrently by threads. 
+--
+-- Any other output lines displayed using
 -- `outputConcurrent` and `createProcessConcurrent`
 -- will scroll up above the open console regions.
 --
@@ -108,6 +110,7 @@ module System.Console.Regions (
 	consoleWidth,
 	consoleHeight,
 	regionList,
+	waitDisplayChange,
 ) where
 
 import Data.Monoid
@@ -207,8 +210,7 @@ consoleWidth = munge . Console.width <$> readTVar consoleSize
 consoleHeight :: STM Int
 consoleHeight = Console.height <$> readTVar consoleSize
 
--- | The RegionList TMVar is left empty when `displayConsoleRegions`
--- is not running.
+-- | Check if `displayConsoleRegions` is running.
 regionDisplayEnabled :: IO Bool
 regionDisplayEnabled = atomically $ not <$> isEmptyTMVar regionList
 
@@ -431,47 +433,93 @@ displayConsoleRegions a = ifM (liftIO regionDisplayEnabled)
 
 trackConsoleWidth :: IO ()
 trackConsoleWidth = do
-	let getwidth = maybe noop (atomically . writeTVar consoleSize)
+	let getsz = maybe noop (atomically . writeTVar consoleSize)
 		=<< Console.size
-	getwidth
-	installResizeHandler (Just getwidth)
+	getsz
+	installResizeHandler (Just getsz)
 
 data DisplayChange
-	= BufferChange (StdHandle, OutputBuffer)
+	= BufferChange BufferSnapshot 
 	| RegionChange RegionSnapshot
+	| RegionListChange RegionSnapshot
 	| TerminalResize Width
-	| EndSignal ()
+	| Shutdown
+	| DisplayChangeBarrier Barrier
 
+type BufferSnapshot = (StdHandle, OutputBuffer)
 type RegionSnapshot = ([ConsoleRegion], [R], [[Text]])
+type Barrier = Integer
+
+-- | This is a broadcast TChan, which gets a DisplayChange written to it
+-- after the display has been updated. It can be used to wait for something
+-- to be displayed.
+{-# NOINLINE displayUpdateNotifier #-}
+displayUpdateNotifier :: TChan DisplayChange
+displayUpdateNotifier = unsafePerformIO $ newBroadcastTChanIO
+
+{-# NOINLINE displayChangeBarrier #-}
+displayChangeBarrier :: TVar Barrier
+displayChangeBarrier = unsafePerformIO $ newTVarIO 0
+
+-- | Runs a STM action, and waits for the display to be fully updated
+-- before returning.
+waitDisplayChange :: STM a -> IO a
+waitDisplayChange a = do
+	r <- atomically a
+	c <- atomically $ dupTChan displayUpdateNotifier
+	b <- atomically $ do
+		!b <- succ <$> readTVar displayChangeBarrier
+		writeTVar displayChangeBarrier b
+		return b
+	atomically $ waitchange c b
+	return r
+  where
+	waitchange c b = do
+		change <- readTChan c
+		case change of
+			DisplayChangeBarrier b' | b' == b -> return ()
+			_ -> waitchange c b
 
 displayThread :: Bool -> TSem -> IO ()
 displayThread isterm endsignal = do
 	origwidth <- atomically consoleWidth
-	go ([], [], []) origwidth
+	origbarrier <- atomically (readTVar displayChangeBarrier)
+	go ([], [], []) origwidth origbarrier
   where
-	go origsnapshot@(orighandles, origregions, origlines) origwidth = do
+	go origsnapshot@(orighandles, origregions, origlines) origwidth origbarrier = do
 		let waitwidthchange = do
 			w <- consoleWidth
 			if w == origwidth then retry else return w
+		let waitbarrierchange = do
+			b <- readTVar displayChangeBarrier
+			if b /= origbarrier
+				then return b
+				else retry
 		let waitanychange =
 			(RegionChange <$> regionWaiter origsnapshot origwidth)
 				`orElse`
-			(RegionChange <$> regionListWaiter origsnapshot)
+			(RegionListChange <$> regionListWaiter origsnapshot)
 				`orElse`
 			(BufferChange <$> outputBufferWaiterSTM waitCompleteLines)
 				`orElse`
 			(TerminalResize <$> waitwidthchange)
 				`orElse`
-			(EndSignal <$> waitTSem endsignal)
+			(waitTSem endsignal >> pure Shutdown)
+				`orElse`
+			-- Must come last, so the changes above are
+			-- processed before barriers.
+			(DisplayChangeBarrier <$> waitbarrierchange)
 		(change, height) <- atomically $ (,)
 			<$> waitanychange
 			<*> consoleHeight
 		let onscreen = take (height - 1) . concat
-		case change of
-			RegionChange snapshot@(_, _, newlines) -> do
-				when isterm $ do
-					changedLines (onscreen origlines) (onscreen newlines)
-				go snapshot origwidth
+		let update snapshot@(_, _, newlines) = do
+			when isterm $
+				changedLines (onscreen origlines) (onscreen newlines)
+			return $ go snapshot origwidth origbarrier
+		next <- case change of
+			RegionChange snapshot -> update snapshot
+			RegionListChange snapshot -> update snapshot
 			BufferChange (h, buf) -> do
 				-- Note that even when every available line
 				-- is dedicated to visible regions, the
@@ -481,13 +529,19 @@ displayThread isterm endsignal = do
 				let origlines' = onscreen origlines
 				inAreaAbove isterm (length origlines') origlines' $
 					emitOutputBuffer h buf
-				go origsnapshot origwidth
+				return $ go origsnapshot origwidth origbarrier
 			TerminalResize newwidth -> do
 				newlines <- atomically (mapM (resizeRegion newwidth) orighandles)
 				when isterm $ do
 					resizeRecovery (onscreen newlines)
-				go (orighandles, origregions, newlines) newwidth
-			EndSignal () -> return ()
+				return $ go (orighandles, origregions, newlines) newwidth origbarrier
+			Shutdown ->
+				return $ return ()
+			DisplayChangeBarrier b ->
+				return $ go origsnapshot origwidth b
+		hFlush stdout
+		atomically $ writeTChan displayUpdateNotifier change
+		next
 
 readRegions :: [ConsoleRegion] -> STM [R]
 readRegions = mapM (\(ConsoleRegion h) -> readTVar h)
@@ -576,7 +630,6 @@ updateLines l
 #endif
 		cursorDown (sum (map (snd . fst) l'))
 		setCursorColumn 0
-		hFlush stdout
   where
 	l' = changeOffsets l 1 []
 
@@ -608,7 +661,6 @@ inAreaAbove isterm numlines ls outputter = do
 	when isterm $ do
 		setCursorColumn 0 -- just in case the output lacked a newline
 		displayLines (reverse ls)
-	hFlush stdout
 
 displayLines :: [Text] -> IO ()
 displayLines = mapM_ $ \l -> do
